/**
 *
 * Copyright (C) Mellanox Technologies Ltd. 2001-2016, 2018.
 * ALL RIGHTS RESERVED.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 */

#include <linux/version.h>
#include <linux/acpi.h>
#include <linux/slab.h>
#include <linux/kref.h>
#include <linux/module.h>
#include <linux/i2c.h>
#include <linux/hwmon.h>
#include <linux/hwmon-sysfs.h>
#include <linux/dmi.h>
#include "mlnx-mux-drv.h"
#include "mlnx-common-drv.h"
#include "mlx_sx1000_sn2000_cpld.h"
#include "platform_defs.h"

#define     DRIVER_VERSION              "1.1"
#define     DRIVER_NAME                 "mlx_1000_2000_cpld"

#define     MLNX_SX_SN_CPLD_BASE      0x2500
#define     MLNX_SX_SN_CPLD_SIZE      0x100

static int num_sys_leds;
static const struct attribute_group *sys_attr_group;
static struct led *sys_cpld_leds;

enum mlnx_system_types
{
    mlnx_dflt_sys_type,
    msn2100_sys_type,
};

const char* mlnx_product_names[] = {
    "DFLT",
    "MSN2100",
    NULL
};

static inline int mlnx_check_system_type(void)
{
    const char* product_name, * mlnx_product_name;
    enum mlnx_system_types mlnx_system_type = mlnx_dflt_sys_type;
    int idx = 1;

    mlnx_product_name = mlnx_product_names[idx];
    product_name = dmi_get_system_info(DMI_PRODUCT_NAME);
    if (product_name) {
        while (mlnx_product_name) {
            if (strstr(product_name, mlnx_product_name)) {
                mlnx_system_type = idx;
                break;
            }
            else {
                mlnx_product_name = mlnx_product_names[++idx];
            }
        }
    }
    return mlnx_system_type;
}

static const char cpld_driver_name[] = DRIVER_NAME;
static struct i2c_client *mlx_sxsn_cpld_client;

static int cpld_rd(u16 offset)
{
    return inb(MLNX_SX_SN_CPLD_BASE + offset);
}

static void cpld_wr(u16 offset, u8 data)
{
    outb(data, MLNX_SX_SN_CPLD_BASE + offset);
}

static u16 cpld_version_offset[] =
{
    CPLD_REG_VERSION1_OFFSET,
    CPLD_REG_VERSION2_OFFSET,
    CPLD_REG_VERSION3_OFFSET
};

static ssize_t cpld_revision_show(struct device * dev,
                                  struct device_attribute * dattr,
                                  char * buf)
{
    int cpld_idx;

    cpld_idx = dattr->attr.name[4] - '1';
    return sprintf(buf, "%u\n", cpld_rd(cpld_version_offset[cpld_idx]));
}
static SYSFS_ATTR_RO(cpld1_revision, cpld_revision_show);
static SYSFS_ATTR_RO(cpld2_revision, cpld_revision_show);
static SYSFS_ATTR_RO(cpld3_revision, cpld_revision_show);


static u8 fan_status_idx[] =
{
    CPLD_FAN1_STATUS, CPLD_FAN2_STATUS,
    CPLD_FAN3_STATUS, CPLD_FAN4_STATUS,
};
static ssize_t fan_status_show(struct device * dev,
                               struct device_attribute * dattr,
                               char * buf)
{
    int fan_idx;
    u8  stat;
    u8  mask;
    u8  bit;

    fan_idx = dattr->attr.name[3] - '1';
    stat = cpld_rd(CPLD_REG_FAN_STATUS_OFFSET);
    mask = cpld_rd(CPLD_REG_FAN_MASK_OFFSET);
    bit = fan_status_idx[fan_idx];
    return sprintf(buf, "%s\n", ((stat & bit) == (mask & bit)) ? "1" : "0");
}
static SYSFS_ATTR_RO(fan1_ok, fan_status_show);
static SYSFS_ATTR_RO(fan2_ok, fan_status_show);
static SYSFS_ATTR_RO(fan3_ok, fan_status_show);
static SYSFS_ATTR_RO(fan4_ok, fan_status_show);


static u16 psu_present_idx[] =
{
    CPLD_PSU1_PRESENT, CPLD_PSU2_PRESENT
};
static ssize_t psu_present_show(struct device * dev,
                               struct device_attribute * dattr,
                               char * buf)
{
    int psu_idx;
    u8  data;

    psu_idx = dattr->attr.name[7] - '1';          // psu_pwrX_present
    data = cpld_rd(CPLD_REG_PSU_PRESENT_OFFSET);
    return sprintf(buf, "%s\n", (data & psu_present_idx[psu_idx]) ? "0" : "1");
}
static SYSFS_ATTR_RO(psu_pwr1_present, psu_present_show);
static SYSFS_ATTR_RO(psu_pwr2_present, psu_present_show);

static u16 psu_ok_idx[] =
{
    CPLD_PSU1_OK, CPLD_PSU2_OK
};
static ssize_t psu_status_show(struct device * dev,
                               struct device_attribute * dattr,
                               char * buf)
{
    int psu_idx;
    u8  data;

    psu_idx = dattr->attr.name[7] - '1';        // psu_pwrX_all_ok
    data = cpld_rd(CPLD_REG_PSU_STATUS_OFFSET);
    return sprintf(buf, "%s\n", (data & psu_ok_idx[psu_idx]) ? "1" : "0");
}
static SYSFS_ATTR_RO(psu_pwr1_all_ok, psu_status_show);
static SYSFS_ATTR_RO(psu_pwr2_all_ok, psu_status_show);

/*------------------------------------------------------------------------------
 *
 * LED definitions
 *
 */

struct led {
	char name[PLATFORM_LED_COLOR_NAME_SIZE];
	uint32_t reg;
	uint8_t mask;
	int n_colors;
	int color_shift;
	struct led_color colors[8];
};

/*
msn2700, msn2410
   status: reg: 0x20 mask: 0x0f
   psu   : reg: 0x20 mask: 0xf0
   fan1  : reg: 0x21 mask: 0x0f
   fan2  : reg: 0x21 mask: 0xf0
   fan3  : rag: 0x22 mask: 0x0f
   fan4  : reg: 0x22 mask: 0xf0

msn2100
   status: reg: 0x20 mask 0x0f
   fan   : reg: 0x21 mask 0x0f
   psu1  : reg: 0x23 mask 0x0f
   psu2  : reg: 0x23 mask 0xf0
   ???   : reg: 0x24 mask 0xf0
*/

static struct led cpld_default_leds[] = {
	{
		.name = "led_status",
		.reg  = CPLD_REG_PSU_SYS_LED,
		.mask = CPLD_LED_MASK,
		.n_colors = 7,
		.color_shift = CPLD_STATUS_LED_SHIFT,
		.colors = {
			{ PLATFORM_LED_GREEN, LED_GREEN_STATIC_ON},
			{ PLATFORM_LED_GREEN_SLOW_BLINKING, LED_GREEN_BLINK_3HZ},
			{ PLATFORM_LED_GREEN_BLINKING, LED_GREEN_BLINK_6HZ},
			{ PLATFORM_LED_RED, LED_RED_STATIC_ON},
			{ PLATFORM_LED_RED_SLOW_BLINKING, LED_RED_BLINK_3HZ},
			{ PLATFORM_LED_RED_BLINKING, LED_RED_BLINK_6HZ},
			{ PLATFORM_LED_OFF, LED_IS_OFF},
		},
	},
	{
		.name = "led_psu",
		.reg  = CPLD_REG_PSU_SYS_LED,
		.mask = CPLD_LED_MASK,
		.n_colors = 7,
		.color_shift = CPLD_PSU_LED_SHIFT,
		.colors = {
			{ PLATFORM_LED_GREEN, LED_GREEN_STATIC_ON},
			{ PLATFORM_LED_GREEN_SLOW_BLINKING, LED_GREEN_BLINK_3HZ},
			{ PLATFORM_LED_GREEN_BLINKING, LED_GREEN_BLINK_6HZ},
			{ PLATFORM_LED_RED, LED_RED_STATIC_ON},
			{ PLATFORM_LED_RED_SLOW_BLINKING, LED_RED_BLINK_3HZ},
			{ PLATFORM_LED_RED_BLINKING, LED_RED_BLINK_6HZ},
			{ PLATFORM_LED_OFF, LED_IS_OFF},
		},
	},
	{
		.name = "led_fan1",
		.reg  = CPLD_REG_FAN1_FAN2_LED,
		.mask = CPLD_LED_MASK,
		.n_colors = 7,
		.color_shift = CPLD_FAN1_LED_SHIFT,
		.colors = {
			{ PLATFORM_LED_GREEN, LED_GREEN_STATIC_ON},
			{ PLATFORM_LED_GREEN_SLOW_BLINKING, LED_GREEN_BLINK_3HZ},
			{ PLATFORM_LED_GREEN_BLINKING, LED_GREEN_BLINK_6HZ},
			{ PLATFORM_LED_RED, LED_RED_STATIC_ON},
			{ PLATFORM_LED_RED_SLOW_BLINKING, LED_RED_BLINK_3HZ},
			{ PLATFORM_LED_RED_BLINKING, LED_RED_BLINK_6HZ},
			{ PLATFORM_LED_OFF, LED_IS_OFF},
		},
	},
	{
		.name = "led_fan2",
		.reg  = CPLD_REG_FAN1_FAN2_LED,
		.mask = CPLD_LED_MASK,
		.n_colors = 7,
		.color_shift = CPLD_FAN2_LED_SHIFT,
		.colors = {
			{ PLATFORM_LED_GREEN, LED_GREEN_STATIC_ON},
			{ PLATFORM_LED_GREEN_SLOW_BLINKING, LED_GREEN_BLINK_3HZ},
			{ PLATFORM_LED_GREEN_BLINKING, LED_GREEN_BLINK_6HZ},
			{ PLATFORM_LED_RED, LED_RED_STATIC_ON},
			{ PLATFORM_LED_RED_SLOW_BLINKING, LED_RED_BLINK_3HZ},
			{ PLATFORM_LED_RED_BLINKING, LED_RED_BLINK_6HZ},
			{ PLATFORM_LED_OFF, LED_IS_OFF},
		},
	},
	{
		.name = "led_fan3",
		.reg  = CPLD_REG_FAN3_FAN4_LED,
		.mask = CPLD_LED_MASK,
		.n_colors = 7,
		.color_shift = CPLD_FAN3_LED_SHIFT,
		.colors = {
			{ PLATFORM_LED_GREEN, LED_GREEN_STATIC_ON},
			{ PLATFORM_LED_GREEN_SLOW_BLINKING, LED_GREEN_BLINK_3HZ},
			{ PLATFORM_LED_GREEN_BLINKING, LED_GREEN_BLINK_6HZ},
			{ PLATFORM_LED_RED, LED_RED_STATIC_ON},
			{ PLATFORM_LED_RED_SLOW_BLINKING, LED_RED_BLINK_3HZ},
			{ PLATFORM_LED_RED_BLINKING, LED_RED_BLINK_6HZ},
			{ PLATFORM_LED_OFF, LED_IS_OFF},
		},
	},
	{
		.name = "led_fan4",
		.reg  = CPLD_REG_FAN3_FAN4_LED,
		.mask = CPLD_LED_MASK,
		.n_colors = 7,
		.color_shift = CPLD_FAN4_LED_SHIFT,
		.colors = {
			{ PLATFORM_LED_GREEN, LED_GREEN_STATIC_ON},
			{ PLATFORM_LED_GREEN_SLOW_BLINKING, LED_GREEN_BLINK_3HZ},
			{ PLATFORM_LED_GREEN_BLINKING, LED_GREEN_BLINK_6HZ},
			{ PLATFORM_LED_RED, LED_RED_STATIC_ON},
			{ PLATFORM_LED_RED_SLOW_BLINKING, LED_RED_BLINK_3HZ},
			{ PLATFORM_LED_RED_BLINKING, LED_RED_BLINK_6HZ},
			{ PLATFORM_LED_OFF, LED_IS_OFF},
		},
	},
};
static int n_default_leds = ARRAY_SIZE(cpld_default_leds);

static struct led cpld_sn2100_leds[] = {
	{
		.name = "led_status",
		.reg  = CPLD_REG_PSU_SYS_LED,
		.mask = CPLD_LED_MASK,
		.n_colors = 7,
		.color_shift = CPLD_STATUS_LED_SHIFT,
		.colors = {
			{ PLATFORM_LED_GREEN, LED_GREEN_STATIC_ON},
			{ PLATFORM_LED_GREEN_SLOW_BLINKING, LED_GREEN_BLINK_3HZ},
			{ PLATFORM_LED_GREEN_BLINKING, LED_GREEN_BLINK_6HZ},
			{ PLATFORM_LED_RED, LED_RED_STATIC_ON},
			{ PLATFORM_LED_RED_SLOW_BLINKING, LED_RED_BLINK_3HZ},
			{ PLATFORM_LED_RED_BLINKING, LED_RED_BLINK_6HZ},
			{ PLATFORM_LED_OFF, LED_IS_OFF},
		},
	},
	{
		.name = "led_psu1",
		.reg  = CPLD_REG_2100_PSU_LED,
		.mask = CPLD_LED_MASK,
		.n_colors = 7,
		.color_shift = CPLD_2100_PSU1_LED_SHIFT,
		.colors = {
			{ PLATFORM_LED_GREEN, LED_GREEN_STATIC_ON},
			{ PLATFORM_LED_GREEN_SLOW_BLINKING, LED_GREEN_BLINK_3HZ},
			{ PLATFORM_LED_GREEN_BLINKING, LED_GREEN_BLINK_6HZ},
			{ PLATFORM_LED_RED, LED_RED_STATIC_ON},
			{ PLATFORM_LED_RED_SLOW_BLINKING, LED_RED_BLINK_3HZ},
			{ PLATFORM_LED_RED_BLINKING, LED_RED_BLINK_6HZ},
			{ PLATFORM_LED_OFF, LED_IS_OFF},
		},
	},
	{
		.name = "led_psu2",
		.reg  = CPLD_REG_2100_PSU_LED,
		.mask = CPLD_LED_MASK,
		.n_colors = 7,
		.color_shift = CPLD_2100_PSU2_LED_SHIFT,
		.colors = {
			{ PLATFORM_LED_GREEN, LED_GREEN_STATIC_ON},
			{ PLATFORM_LED_GREEN_SLOW_BLINKING, LED_GREEN_BLINK_3HZ},
			{ PLATFORM_LED_GREEN_BLINKING, LED_GREEN_BLINK_6HZ},
			{ PLATFORM_LED_RED, LED_RED_STATIC_ON},
			{ PLATFORM_LED_RED_SLOW_BLINKING, LED_RED_BLINK_3HZ},
			{ PLATFORM_LED_RED_BLINKING, LED_RED_BLINK_6HZ},
			{ PLATFORM_LED_OFF, LED_IS_OFF},
		},
	},
	{
		.name = "led_fan",
		.reg  = CPLD_REG_FAN1_FAN2_LED,
		.mask = CPLD_LED_MASK,
		.n_colors = 7,
		.color_shift = CPLD_FAN1_LED_SHIFT,
		.colors = {
			{ PLATFORM_LED_GREEN, LED_GREEN_STATIC_ON},
			{ PLATFORM_LED_GREEN_SLOW_BLINKING, LED_GREEN_BLINK_3HZ},
			{ PLATFORM_LED_GREEN_BLINKING, LED_GREEN_BLINK_6HZ},
			{ PLATFORM_LED_RED, LED_RED_STATIC_ON},
			{ PLATFORM_LED_RED_SLOW_BLINKING, LED_RED_BLINK_3HZ},
			{ PLATFORM_LED_RED_BLINKING, LED_RED_BLINK_6HZ},
			{ PLATFORM_LED_OFF, LED_IS_OFF},
		},
	},
};
static int n_sn2100_leds = ARRAY_SIZE(cpld_sn2100_leds);

/*
 * LEDs
 */
static ssize_t led_show(struct device * dev,
                        struct device_attribute * dattr,
                        char * buf)
{
	uint8_t tmp;
	int i;
	struct led * target = NULL;

	/* find the target led */
	for (i = 0; i < num_sys_leds; i++) {
		if (strcmp(dattr->attr.name, sys_cpld_leds[i].name) == 0) {
			target = &sys_cpld_leds[i];
			break;
		}
	}
	if (target == NULL) {
		return sprintf(buf, "undefined target\n");
	}

	/* read the register */
	tmp = cpld_rd(target->reg);
	tmp >>= target->color_shift;

	/* find the color */
	tmp &= target->mask;
	for (i = 0; i < target->n_colors; i++) {
		if (tmp == target->colors[i].value) {
			break;
		}
	}
	if (i == target->n_colors) {
		return sprintf(buf, "undefined color\n");
	} else {
		return sprintf(buf, "%s\n", target->colors[i].name);
	}
}

static ssize_t led_store(struct device * dev,
                         struct device_attribute * dattr,
                         const char * buf, size_t count)
{
	uint8_t tmp;
	int i;
	struct led * target = NULL;
	char raw[PLATFORM_LED_COLOR_NAME_SIZE];

	/* find the target led */
	for (i = 0; i < num_sys_leds; i++) {
		if (strcmp(dattr->attr.name, sys_cpld_leds[i].name) == 0) {
			target = &sys_cpld_leds[i];
			break;
		}
	}
	if (target == NULL) {
		return -EINVAL;
	}

	/* find the color */
	if (sscanf(buf, "%19s", raw) <= 0) {
		return -EINVAL;
	}
	for (i = 0; i < target->n_colors; i++) {
		if (strcmp(raw, target->colors[i].name) == 0) {
			break;
		}
	}
	if (i == target->n_colors) {
		return -EINVAL;
	}

	/* set the new value */
	tmp = cpld_rd(target->reg);
	tmp &= ~(target->mask << target->color_shift);
	tmp |= (target->colors[i].value << target->color_shift);
	cpld_wr(target->reg, tmp);

	return count;
}
/* common LED */
static SYSFS_ATTR_RW(led_status, led_show, led_store);
/* default LEDs */
static SYSFS_ATTR_RW(led_psu, led_show, led_store);
static SYSFS_ATTR_RW(led_fan1, led_show, led_store);
static SYSFS_ATTR_RW(led_fan2, led_show, led_store);
static SYSFS_ATTR_RW(led_fan3, led_show, led_store);
static SYSFS_ATTR_RW(led_fan4, led_show, led_store);
/* msn2100 LEDs */
static SYSFS_ATTR_RW(led_fan, led_show, led_store);
static SYSFS_ATTR_RW(led_psu1, led_show, led_store);
static SYSFS_ATTR_RW(led_psu2, led_show, led_store);

static ssize_t reset_cause_show(struct device * dev,
                                struct device_attribute * dattr,
                                char * buf)
{
    u8  data;
    const char *cause;

    data = cpld_rd(CPLD_REG_RESET_CAUSE_OFFSET);
    switch (data) {
        case CPLD_RESET_CAUSE_CLEAN:
            cause = "clean";
            break;
        case CPLD_RESET_CAUSE_LONG_PB:
            cause = "long press button";
            break;
        case CPLD_RESET_CAUSE_SHORT_PB:
            cause = "short press button";
            break;
        case CPLD_RESET_CAUSE_AUXPWR_OR_FU:
            cause = "aux pwr off or field upgrade";
            break;
        case CPLD_RESET_CAUSE_POWER_FAIL:
            cause = "power fail";
            break;
        case CPLD_RESET_CAUSE_SW_RST:
            cause = "sw reset";
            break;
        case CPLD_RESET_CAUSE_FW_RST:
            cause = "fw reset";
            break;
        case CPLD_RESET_CAUSE_WATCHDOG:
            cause = "watch dog";
            break;
        case CPLD_RESET_CAUSE_THERMAL:
            cause = "switch board power fail";
            break;
        default:
            cause = "unknown or mixed";
            break;
    }
    return sprintf(buf, "%s", cause);
}
static SYSFS_ATTR_RO(reset_cause, reset_cause_show);

/*------------------------------------------------------------------------------
 *
 * sysfs registration
 *
 */

static struct attribute *default_cpld_attrs[] = {
    &dev_attr_cpld1_revision.attr,
    &dev_attr_cpld2_revision.attr,
    &dev_attr_cpld3_revision.attr,
    &dev_attr_fan1_ok.attr,
    &dev_attr_fan2_ok.attr,
    &dev_attr_fan3_ok.attr,
    &dev_attr_fan4_ok.attr,
    &dev_attr_psu_pwr1_present.attr,
    &dev_attr_psu_pwr2_present.attr,
    &dev_attr_psu_pwr1_all_ok.attr,
    &dev_attr_psu_pwr2_all_ok.attr,
    &dev_attr_led_status.attr,
    &dev_attr_led_psu.attr,
    &dev_attr_led_fan1.attr,
    &dev_attr_led_fan2.attr,
    &dev_attr_led_fan3.attr,
    &dev_attr_led_fan4.attr,
    &dev_attr_reset_cause.attr,
    NULL,
};

static struct attribute_group default_cpld_attr_group = {
	.attrs = default_cpld_attrs,
};

static struct attribute *sn2100_cpld_attrs[] = {
    &dev_attr_cpld1_revision.attr,
    &dev_attr_cpld2_revision.attr,
    &dev_attr_cpld3_revision.attr,
    &dev_attr_fan1_ok.attr,
    &dev_attr_fan2_ok.attr,
    &dev_attr_fan3_ok.attr,
    &dev_attr_fan4_ok.attr,
    &dev_attr_psu_pwr1_all_ok.attr,
    &dev_attr_psu_pwr2_all_ok.attr,
    &dev_attr_led_status.attr,
    &dev_attr_led_psu1.attr,
    &dev_attr_led_psu2.attr,
    &dev_attr_led_fan.attr,
    &dev_attr_reset_cause.attr,
    NULL,
};

static struct attribute_group sn2100_cpld_attr_group = {
	.attrs = sn2100_cpld_attrs,
};

/*------------------------------------------------------------------------------
 *
 * CPLD driver interface
 *
 */
static int mlx_sxsn_cpld_probe(struct i2c_client* client,
                                        const struct i2c_device_id* id)
{
    int retval = 0;
    const char *product_name;
    struct kobject * kobj = &client->dev.kobj;

    if (dev_get_drvdata(&client->dev)) {
        printk(KERN_ERR "already probed\n");
        return 0;
    }
    mlx_sxsn_cpld_client = client;

    sys_attr_group = &default_cpld_attr_group;
    num_sys_leds = n_default_leds;
    sys_cpld_leds = cpld_default_leds;
    product_name = dmi_get_system_info(DMI_PRODUCT_NAME);

    if (strstr(product_name, "MSN2010")) {
        int i;

        /*
         * Apparently, the MSN2010 differs from the default attribute model
         * only in the shift of the color bits for led_psu.  Chop that shift
         * from 4 bits to 0 and they are identical.  So rather than dup all
         * the attribute structures and change *one* for 2010, we simply
         * patch color_shift to 0 here once.
         *
         * This will go away when we update our Mellanox drivers to current
         * (which will contain support for the very new 2010).
         */
	for (i = 0; i < num_sys_leds; i++) {
		if (strcmp(sys_cpld_leds[i].name, "led_psu") == 0) {
			struct led *target;

			target = &sys_cpld_leds[i];
			target->color_shift = 0;
			break;
		}
	}
    }

    if (msn2100_sys_type == mlnx_check_system_type()) {
        sys_attr_group = &sn2100_cpld_attr_group;
        num_sys_leds = n_sn2100_leds;
        sys_cpld_leds = cpld_sn2100_leds;
    }

    retval = sysfs_create_group(kobj, sys_attr_group);
    if (retval) {
        return retval;
    }
    return 0;
}

static int mlx_sxsn_cpld_remove(struct i2c_client *client)
{
    struct kobject * kobj = &client->dev.kobj;

    sysfs_remove_group(kobj, sys_attr_group);
    return 0;
}

static struct i2c_device_id mlx_sxsn_cpld_ids[] = {
	{
		.name = DRIVER_NAME,
	},
	{ /* end of list */ },
};
MODULE_DEVICE_TABLE(i2c, mlx_sxsn_cpld_ids);

static struct i2c_driver mlx_sxsn_cpld_driver = {
	.probe = mlx_sxsn_cpld_probe,
	.remove = mlx_sxsn_cpld_remove,
	.id_table = mlx_sxsn_cpld_ids,
	.driver = {
		.name  = cpld_driver_name,
		.owner = THIS_MODULE,
	},
};


/*------------------------------------------------------------------------------
 *
 * module interface
 *
 */

static int __init mlx_sxsn_cpld_init(void)
{
	int rv;

	rv = i2c_add_driver(&mlx_sxsn_cpld_driver);
	if (rv) {
		printk(KERN_ERR
		       "%s i2c_add_driver() failed (%i)\n",
		       cpld_driver_name, rv);
		goto exit;
	}
	pr_info(DRIVER_NAME": version "DRIVER_VERSION" successfully loaded\n");

exit:
	return rv;
}

static void __exit mlx_sxsn_cpld_exit(void)
{
	i2c_del_driver(&mlx_sxsn_cpld_driver);
}

MODULE_AUTHOR("Alan Liebthal <alanl@cumulusnetworks.com>");
MODULE_DESCRIPTION("Mellanox SX/SN CPLD Support");
MODULE_LICENSE("GPL");
MODULE_VERSION(DRIVER_VERSION);

module_init(mlx_sxsn_cpld_init);
module_exit(mlx_sxsn_cpld_exit);
